package sysModel.fish;

import model.ILambda;
import model.RandNumGenerator;
import sysModel.env.AGlobalEnv;
import sysModel.env.FishApplyParams;

import java.awt.*;
import java.util.Observable;
import java.util.Observer;

/**
 * Behavior-controlled fish. *
 *
 * @author Mathias G. Ricken
 */
public class Fish implements Cloneable, Observer {
    /**
     * The local environment of the fish.
     */
    private AGlobalEnv.ALocalEnv _localEnv;

    /**
     * The probability that this fish attempts to breed this timestep. Subclasses can override this value.
     */
    protected double _probOfBreeding = 0; // 1.0f/7.0f;

    /**
     * The probability that this fish will die in this timestep. Subclasses can override this value.
     */
    protected double _probOfDying = 0; // 1.0f/5.0f;

    /**
     * Color of the fish.
     */
    protected Color _fishColor;

    /**
     * Display for the fish.
     */
    protected IFishDisplay _fishDisplay;

    /**
     * Behavior for the fish.
     */
    protected IFishBehavior _fishBehavior;

    /**
     * Create a new fish.
     *
     * @param fishColor    fish color
     * @param fishDisplay  fish display
     * @param fishBehavior fish behavior
     */
    public Fish(Color fishColor, IFishDisplay fishDisplay, IFishBehavior fishBehavior) {
        _fishColor = fishColor;
        _fishDisplay = fishDisplay;
        _fishBehavior = fishBehavior;
    }

    /**
     * Set the fish behavior.
     *
     * @param fishBehavior new fish behavior.
     */
    public void setFishBehavior(IFishBehavior fishBehavior) {
        _fishBehavior = fishBehavior;
    }

    /**
     * Get the fish behavior.
     *
     * @return fish behavior
     */
    public IFishBehavior getFishBehavior() {
        return _fishBehavior;
    }

    /**
     * Set the local environment of this fish.
     *
     * @param localEnv local environment
     */
    public final void setLocalEnvironment(AGlobalEnv.ALocalEnv localEnv) {
        _localEnv = localEnv;
    }

    /**
     * Attempt to move the fish forward, which may or may not be successful. The behavior in each case is defined by the
     * visitor: - If the move cannot be executed, the blockedCmd lambda is applied. The parameter is not used and set to
     * null. - If the move can be executed, the openCmd lambda is applied. The parameter is an ILambda that can to be
     * executed to actually move the fish to the target location of this move. The ILambda ignores the input parameter
     * and returns null.
     *
     * @param blockedCmd lambda to apply if blocked
     * @param openCmd    lambda to apply if open
     *
     * @return return value of lambda executed
     */
    public final Object tryMoveFwd(final ILambda blockedCmd, final ILambda openCmd) {
        return _localEnv.tryMoveFwd(this, blockedCmd, openCmd);
    }

    /**
     * Attempt to breed the fish forward, which may or may not be successful. The behavior in each case is defined by
     * the visitor: - If the breeding cannot be executed, the blockedCmd lambda is applied. The parameter is not used
     * and set to null. - If the breeding can be executed, the openCmd lambda is applied. The parameter is an ILambda
     * that can to be executed to actually move the fish to the target location of this breeding. The ILambda ignores
     * the input parameter and returns null.
     *
     * @param blockedCmd lambda to apply if blocked
     * @param openCmd    lambda to apply if open
     *
     * @return return value of lambda executed
     */
    public final Object tryBreedFwd(final ILambda blockedCmd, final ILambda openCmd) {
        return _localEnv.tryBreedFwd(this, blockedCmd, openCmd);
    }

    /**
     * Draw the fish on the graphics object. The graphics object still has to be translated and rotated properly,
     *
     * @param g    graphics object to drawFish on
     * @param comp component to drawFish on
     */
    public final void draw(Graphics2D g, Component comp) {
        _localEnv.drawFish(this, g, comp);
    }

    /**
     * Turn the fish radians to the right.
     *
     * @param radians radians to turn
     */
    public final void turnRight(double radians) {
        _localEnv.turnRight(this, radians);
    }

    /**
     * Turn the fish radians to the left.
     *
     * @param radians radians to turn
     */
    public final void turnLeft(double radians) {
        _localEnv.turnRight(this, -radians);
    }

    /**
     * Turn the fish Pi/2 radians to the right.
     */
    public final void turnRight() {
        turnRight(Math.PI / 2);
    }

    /**
     * Turn the fish Pi/2 radians to the left.
     */
    public final void turnLeft() {
        turnLeft(Math.PI / 2);
    }

    /**
     * Let the fish die, remove it from the environment and the simulation.
     */
    public final void die() {
        _localEnv.removeFish(this);
    }

    /**
     * Execute a simulation step. This function acts as a template method. It always calls the breed, move and age
     * functions, which can be overridden, though.
     */
    public final void act() {
        breed();
        move();
        age();
    }

    /**
     * Execute the breeding part of a simulation step.
     */
    protected void breed() {
        if (RandNumGenerator.instance().nextDouble() < _probOfBreeding) {
            // breed
            // by default, a fish breeds into all open spaces around it
            final ILambda blocked1 = new ILambda() {
                public Object apply(Object param) {
                    // turn PI/2 radians to the right to face in the original direction
                    turnRight();
                    return null;
                }
            };

            final ILambda blocked2 = new ILambda() {
                public Object apply(Object param) {
                    // turn PI/2 radians to the right and attempt to breed forward
                    turnRight();
                    return tryBreedFwd(blocked1, new ILambda() {
                        public Object apply(Object param) {
                            // the field to the right is open
                            ((ILambda)param).apply(null);
                            return blocked1.apply(null);
                        }
                    });
                }
            };

            final ILambda blocked3 = new ILambda() {
                public Object apply(Object param) {
                    // turn PI/2 radians to the right and attempt to breed forward
                    turnRight();
                    return tryBreedFwd(blocked2, new ILambda() {
                        public Object apply(Object param) {
                            // the field to the right is open
                            ((ILambda)param).apply(null);
                            return blocked2.apply(null);
                        }
                    });
                }
            };

            final ILambda blocked4 = new ILambda() {
                public Object apply(Object param) {
                    // turn PI/2 radians to the right and attempt to breed forward
                    turnRight();
                    return tryBreedFwd(blocked3, new ILambda() {
                        public Object apply(Object param) {
                            // the field straight ahead is open
                            ((ILambda)param).apply(null);
                            return blocked3.apply(null);
                        }
                    });
                }
            };

            // attempt to breed forward
            tryBreedFwd(blocked4, new ILambda() {
                public Object apply(Object param) {
                    // the field to the left is open
                    ((ILambda)param).apply(null);
                    return blocked4.apply(null);
                }
            });
        }
    }

    /**
     * Execute the aging part of a simulation step.
     */
    protected void age() {
        if (RandNumGenerator.instance().nextDouble() < _probOfDying) {
            die();
        }
    }

    /**
     * Return a clone this instance.
     *
     * @return clone
     */
    public Object clone() {
        Object clone;
        try {
            clone = super.clone();
        }
        catch(CloneNotSupportedException e) {
            throw new FishException("Fish.clone failed", e);
        }
        return clone;
    }

    /**
     * Return a string representation of this fish.
     *
     * @return string representation
     */
    public String toString() {
        String className = this.getClass().getName();
        return className.substring(className.lastIndexOf('.') + 1);
    }

    /**
     * Execute the movement part of a simulation step. PART4
     */
    protected void move() {
        _fishBehavior.move(this);
    }

    /**
     * Draw the fish. The Graphics2D object has been translated, rotated and scaled so that the origin is in the center
     * of the fish. The fish should just be drawn from (-0.5, -0.5) to (0.5, 0.5).
     *
     * @param g    graphics object to drawFish on.
     * @param comp the component to drawFish on
     */
    public void paint(Graphics2D g, Component comp) {
        _fishDisplay.draw(g, comp, getColor());
    }

    /**
     * Get the fish's color.
     *
     * @return color of the fish
     */
    public Color getColor() {
        return _fishColor;
    }

    /**
     * This method is called whenever the observed object is changed. The fish executes the lambda that is passed in
     * with the fish itself as parameter.
     *
     * @param o   the observable object
     * @param arg lambda to execute
     */
    public final void update(final Observable o, Object arg) {
        try {
            ((ILambda)arg).apply(new FishApplyParams(this, _localEnv));
        }
        catch(Throwable t) {
            // let fish die
            die();
            System.err.println("Fish throws " + t);
            t.printStackTrace();
            System.gc();
            // System.err.println("\tPostGC\tTotal memory: "+Runtime.getRuntime().totalMemory()+", Free memory: "+Runtime.getRuntime().freeMemory());
            // System.err.println("-------------------------");
        }
    }

    /**
     * Set probability of dying.
     *
     * @param p the probability
     */
    public void setProbOfDying(double p) {
        _probOfDying = p;
    }

    /**
     * Set probability of breeding.
     *
     * @param p the probability
     */
    public void setProbOfBreeding(double p) {
        _probOfBreeding = p;
    }
}
